/* sqfig.c -- Square-Figure functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqfig.c,v 1.2 2003/03/17 09:41:35 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/

/* Note: the memory policy is not to free anything that might still be
* referenced.
*/

/* The functions sqalloc(), sqrealloc() and sqfree() have the same
* interface the well known functions malloc(), realloc() and free()
* have, with the exception that they don't return on out-of-memory
* conditions [i.e. call abort() instead].
*/


/* implemented interface */
#include "sqfig.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* local stuff */
static void addFigureToFigureSet(
	figureSet_t* figureSet,
	figure_t const* figure);
static void addSquareToFigure(
	figure_t* destination,
	figure_t const* source,
	square_t* const square,
	square_t* const hint);
static void addSquareToFigureSet(
	figureSet_t* destination,
	figureSet_t const* source);
static void* bsearchx(
	void const* key,
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b),
	void** position);
static figureSet_t* createEmptyFigureSet(void);
static figureSet_t* createFigureSet(int size);
static figureSet_t* createOneSquareFigureSet(void);
static int figureCompare(figure_t const* a, figure_t const* b);
static void expandFigureSetCache(int size);
static int figurecmp1(void const* a, void const* b);
static int figurecmp2(void const* a, void const* b);
static figure_t* findFigure2(
	figure_t** figures,
	int from,
	int to,
	figure_t const* what,
	figure_t*** position);
static siblingGroup_t* findSiblingGroup2(
	siblingGroup_t** siblingGroups,
	int from,
	int to,
	siblingGroup_t const* what,
	siblingGroup_t*** position);
static square_t* findSquare1(
	square_t* squares,
	int from,
	int to,
	square_t const* what,
	square_t** position);
static figureList_t* generateFlattened(siblingList_t const* siblings);
static figureSet_t* generateNextFigureSet(figureSet_t const* figureSet);
static siblingGroup_t* generateSiblingGroup(figure_t const* figure);
static void mirrorFigure(figure_t* destination, figure_t const* source);
static void normalizeFigure(figure_t* figure);
static void normalizeFigureList(figureList_t* figureList);
static void normalizeSiblingGroup(siblingGroup_t* siblingGroup);
static void normalizeSiblingList(siblingList_t* siblingList);
static void rotateFigure(figure_t* destination, figure_t const* source);
static void* search(
	void const* key,
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b),
	void** position);
static int siblinggroupcmp2(void const* a, void const* b);
static void sort(
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b));
static void* sqalloc(int size);
static void sqfree(void* p);
static void* sqrealloc(void* p, int size);
static void* sqreallocExponential(void* p, int size);
static int squarecmp1(void const* a, void const* b);


/** Cache for figure set entries, indexed by figure size.  An array
* instead of a (more sparse) map, since the enries are typically
* calculated consecutively anyway.
*/
static figureSet_t** figureSetCache = NULL;
static int figureSetCacheSize = 0;


/** Return a figure set; create it if necessary.  Indirect recursive.
*/
figureSet_t const* getFigureSet(int size)
{
	figureSet_t* fs;

	/* See if the figure set is already in the cache,
	* make sure it's large enough first, and add
	* entry afterwards.
	*/

	expandFigureSetCache(size + 1);

	fs = figureSetCache[size];
	if (fs == NULL) {
		fs = createFigureSet(size);
		figureSetCache[size] = fs;
	}

	/* constness: we don't use const internal,
	* so we'll have to cast here.
	*/
	return (figureSet_t const*)fs;
}


/** Initialize and expand figure set cache.
*/
static void expandFigureSetCache(int size)
{
	int oldSize, i;

	if (figureSetCache != NULL && figureSetCacheSize >= size) {
		return;
	}

	/* reallocation strategy: double size (consider zero) */
	oldSize = figureSetCacheSize;
	while (figureSetCacheSize < size) {
		if (figureSetCacheSize < 1) {
			figureSetCacheSize = 1;
		} else {
			figureSetCacheSize *= 2;
		}
	}

	figureSetCache = sqrealloc(
		figureSetCache,
		figureSetCacheSize * sizeof(*figureSetCache));

	/* Null-out new entries. */
	for (i = oldSize; i < figureSetCacheSize; i++) {
		figureSetCache[i] = NULL;
	}
}


/** Create (calculate) a figure set.  Indirect recursive.
*/
static figureSet_t* createFigureSet(int size)
{
	if (size < 0) {
		/* "can't have FigureSet of negative size" */
		/* errno = EINVAL; */
		return NULL;
	}

	/* small special cases (also: anchors of the recursion) */
	switch (size) {
		
	case 0 :

		/* empty set */
		return createEmptyFigureSet();

	case 1 :

		return createOneSquareFigureSet();
	}

	/* (indirect) recursion step */
	return generateNextFigureSet(getFigureSet(size - 1));
}


/** Create empty figure set.
*/
static figureSet_t* createEmptyFigureSet(void)
{
	figureSet_t* fs;

	fs = sqalloc(sizeof(*fs));

	/* empty array of sibling sets */
	fs->siblings = sqalloc(sizeof(*fs->siblings));
	fs->siblings->size = 0;
	fs->siblings->siblings = NULL;

	/* empty array of figures */
	fs->figures = sqalloc(sizeof(*fs->figures));
	fs->figures->size = 0;
	fs->figures->figures = NULL;

	return fs;
}


/** Create one square figure set.
*/
static figureSet_t* createOneSquareFigureSet(void)
{
	figureSet_t* fs;

	fs = sqalloc(sizeof(*fs));

	/* array of sibling sets, one entry */
	fs->siblings = sqalloc(sizeof(*fs->siblings));
	fs->siblings->size = 1;
	fs->siblings->siblings = sqalloc(
		fs->siblings->size * sizeof(*fs->siblings->siblings));

	/* one figure */
	fs->siblings->siblings[0] = sqalloc(sizeof(*fs->siblings->siblings[0]));
	fs->siblings->siblings[0]->size = 1;

	/* figure with one square */
	fs->siblings->siblings[0]->figures[0].size = 1;

	/* figure with one square, default square is {0,0} */
	fs->siblings->siblings[0]->figures[0].squares[0].x = 0;
	fs->siblings->siblings[0]->figures[0].squares[0].y = 0;

	/* flattened version */
	/* This is very simple since there is only one entry (which obviosly
	* does not need to be sorted).
	*/

	/* array with one figure */
	fs->figures = sqalloc(sizeof(*fs->figures));
	fs->figures->size = 1;
	fs->figures->figures = sqalloc(
		fs->figures->size * sizeof(*fs->figures->figures));

	fs->figures->figures[0] = &fs->siblings->siblings[0]->figures[0];

	return fs;
}


/** Generate a new figure set by adding a square to a smaller figure set
* in every possible way.
*/
static figureSet_t* generateNextFigureSet(figureSet_t const* figureSet)
{
	figureSet_t* nset;

	nset = createEmptyFigureSet();
	addSquareToFigureSet(nset, figureSet);

	return nset;
}


/** Add a square to a figure set in every possible way.
*/
static void addSquareToFigureSet(
	figureSet_t* destination,
	figureSet_t const* source)
{
	int isize, i, jsize, j, k;
	siblingGroup_t** sgs;
	figure_t *f, nf;
	square_t *sqs, *s, ns, *t;

	/* Iterate over the _distinct_ figures.  Take the first entry in the
	* sibling group, the rest isn't needed since the obtainable figures 
	* are created by rotation and mirroring, as required by the hierarchical
	* data structures.
	*/
	isize = source->siblings->size;
	sgs = source->siblings->siblings;
	for (i = 0; i < isize; i++) {
		/* Take the first figure.  There is always at least one figure. */
		f = &sgs[i]->figures[0];

		/* iterate over the squares */
		jsize = f->size;
		sqs = f->squares;
		for (j = 0; j < jsize; j++) {
			s = &sqs[j];

			/* iterate over the square's neighbors */
			for (k = 0; k < 4; k++) {

				SQUARE_COPY(&ns, s);
				t = NULL;

				/* see if the neighbor square is already part of the figure,
				* skip isf so
				*/

				/* Note: the test for findSquare1(sqs,0,jsize,&ns)!=NULL
				* would be sufficient, but the used approach may be more
				* performant.
				*/
				/* Note: a switch statement would have been more performant
				* but doesn't as easy allow to use continue statements to skip
				* to the next loop iteration.
				*/

				if (k == 0) { /* x += 1 */
					ns.x++;

					/* must be @j+1, so just check next;
					* [j+1] would also be the insert position,
					* (sqs[j+1] is valid data anyway)
					*/
					t = &sqs[j + 1];
					if (j + 1 < jsize) {
						if (SQUARE_EQUAL(t, &ns)) {
							continue;
						}
					}

				} else if (k == 1) { /* x -= 1 */
					ns.x--;

					/* must be @j-1, so just check previous */
					if (j > 0) {
						/* [using temp variable since macro multiply evaluates */
						t = &sqs[j - 1];
						if (SQUARE_EQUAL(t, &ns)) {
							continue;
						}
					}

					/* insert position is @j */
					t = &sqs[j];

				} else if (k == 2) { /* y += 1 */
					ns.y++;

					/* must be after j, so juct search from here */
					if (findSquare1(sqs, j + 1, jsize, &ns, &t) != NULL) {
						continue;
					}

				} else if (k == 3) { /* y -= 1 */
					ns.y--;

					/* must be before j, so just find up to here */
					if (findSquare1(sqs, 0, j, &ns, &t) != NULL) {
						continue;
					}
				}

				/* add the new square */
				addSquareToFigure(&nf, f, &ns, t);

				/* See if we already have this figure, skip if so.
				* Note that we don't need findFigure2()'s position hint
				* since the figure will not be inserted in that search
				* array but in the siblings array.
				*/
				if (
					findFigure2(
						destination->figures->figures,
						0,
						destination->figures->size,
						&nf,
						NULL) != NULL
				) {
					continue;
				}

				/* add to the set */
				addFigureToFigureSet(destination, &nf);
			}
		}
	}
}


/** Add a specific square to a figure.
*/
static void addSquareToFigure(
	figure_t* destination,
	figure_t const* source,
	square_t* const square,
	square_t* const hint)
{
	int i = -1, j;

	if (source->size >= MAXFIGURESIZE) {
		abort();
	}

	/* copy */
	memcpy(destination, source, sizeof(*destination));

	/* consider pre-calculated position hint */
	if (hint != NULL) {
		i = hint - source->squares;
		/* sanity check [not really necessary...] */
		if (i < 0 || i > source->size) {
			i = -1;
		}
	}

	if (i != -1) {
		memmove(
			&destination->squares[i + 1],
			&destination->squares[i],
			(source->size - i) * sizeof(*destination->squares));
	}

	/* add */
	j = (i == -1 ? source->size : i);
	SQUARE_COPY(&destination->squares[j], square);
	destination->size++;

	/* normalize */
	if (
		/* unhinted figure needs to be fully normalized */
		i == -1 ||
		/* needs to be offset-normalized; use full normalization too */
		(square->x < 0 || square->y < 0)
	) {
		normalizeFigure(destination);
	}
}


/** Add a figure to a figure set.
*/
static void addFigureToFigureSet(
	figureSet_t* figureSet,
	figure_t const* figure)
{
	siblingGroup_t *sg, **p;
	int i, j;
	figure_t** t;

	sg = generateSiblingGroup(figure);

	/* Inserting the new sibling group in at the right place
	* instead of calling normalizeSiblingList(figureSet->siblings)
	* many times will be a bit faster.
	*/

	figureSet->siblings->siblings = sqreallocExponential(
		figureSet->siblings->siblings,
		(figureSet->siblings->size + 1) *
			sizeof(*figureSet->siblings->siblings));

	findSiblingGroup2(
		figureSet->siblings->siblings,
		0,
		figureSet->siblings->size,
		sg,
		&p);
	j = p - figureSet->siblings->siblings;
	memmove(
		&figureSet->siblings->siblings[j + 1],
		&figureSet->siblings->siblings[j],
		(figureSet->siblings->size - j) * sizeof(*figureSet->siblings->siblings));

	figureSet->siblings->siblings[j] = sg;
	figureSet->siblings->size++;

	/* Inserting the (up to 8) new figures in an already existent
	* flattened list instead of recreating the whole thing with
	* generateFlattened() will be a bit faster.
	*/

	figureSet->figures->figures = sqreallocExponential(
		figureSet->figures->figures,
		(figureSet->figures->size + sg->size) *
			sizeof(*figureSet->figures->figures));

	for (i = 0; i < sg->size; i++) {
		findFigure2(
			figureSet->figures->figures,
			0,
			figureSet->figures->size,
			&sg->figures[i],
			&t);
		j = t - figureSet->figures->figures;
		memmove(
			&figureSet->figures->figures[j + 1],
			&figureSet->figures->figures[j],
			(figureSet->figures->size - j) * sizeof(*figureSet->figures->figures));
		figureSet->figures->figures[j] = &sg->figures[i];
		figureSet->figures->size++;
	}
}


/** Generate a sibling group from a figure.
*/
static siblingGroup_t* generateSiblingGroup(figure_t const* figure)
{
	siblingGroup_t* sg;
	figure_t f, g, m;
	int i;
	int me;

	sg = sqalloc(sizeof(*sg));
	sg->size = 0;

	/* start up with us */
	FIGURE_COPY(&sg->figures[sg->size], figure);
	sg->size++;

	/* Rotate until back to identity.
	* This will take 1, 2, or 4 steps,
	* depending on the figure's symmetry.
	*/
	rotateFigure(&f, figure);
	while (!FIGURE_EQUAL(&f, figure)) {
		/* add sibling */
		FIGURE_COPY(&sg->figures[sg->size], &f);
		sg->size++;

		FIGURE_COPY(&g, &f);
		rotateFigure(&f, &g);
	}

	/* Get the mirror and check if it's already
	* one of the siblings.  This will depend
	* on the figure's symmetry.  Note that all
	* equalities are likeable: if symmetry is along
	* the X axis, #1 will be equal, if along Y it
	* will be #3, if along the diagonals it will be
	* #0 or #2.  We can't use findFigure1() since
	* the array is not yet sorted, which is however
	* not really necessary since only up to 4
	* comparisions are necessary.
	*/
	mirrorFigure(&m, figure);
	me = 0;
	for (i = 0; i < sg->size; i++) {
		if (FIGURE_EQUAL(&m, &sg->figures[i])) {
			me = 1;
			break;
		}
	}
	if (!me) {
		/* Add mirror and its rotations
		* [explanations: see comments above].
		*/

		FIGURE_COPY(&sg->figures[sg->size], &m);
		sg->size++;

		rotateFigure(&f, &m);
		while (!FIGURE_EQUAL(&f, &m)) {
			FIGURE_COPY(&sg->figures[sg->size], &f);
			sg->size++;

			FIGURE_COPY(&g, &f);
			rotateFigure(&f, &g);
		}
	}

	/* normalize (i.e. order (i.e. sort)) */
	normalizeSiblingGroup(sg);

	return sg;
}


/** Rotate and normalize a figure.
*/
static void rotateFigure(figure_t* destination, figure_t const* source)
{
	int i;

	/* Note: X could be directly normalized by maxPreY, but we leave it
	* to normalizeFigure().  Its sort step will be more expensive anyway.
	*/

	destination->size = source->size;
	for (i = 0; i < source->size; i++) {
		/* rotate left: y' := x ; x' := -y */
		destination->squares[i].x = (squareCoordinate_t)-source->squares[i].y;
		destination->squares[i].y = source->squares[i].x;
	}

	normalizeFigure(destination);
}


/** Mirror a figure.
*/
static void mirrorFigure(figure_t* destination, figure_t const* source)
{
	int i;

	/* Note: the coordinates will be already offset-normalized,
	* normalizeFigure() still needs to sort (which is more expensive).
	*/

	destination->size = source->size;
	for (i = 0; i < source->size; i++) {
		/* flip x and y: y' := x ; x' := y */
		destination->squares[i].x = source->squares[i].y;
		destination->squares[i].y = source->squares[i].x;
	}

	normalizeFigure(destination);
}


/** Return a flattened version of the hierarchical siblings data
* structure.
*/
static figureList_t* generateFlattened(siblingList_t const* siblings)
{
	int n, i, jsize, j;
	figureList_t* fl;
	figure_t* f;

	/* count entries */
	n = 0;
	for (i = 0; i < siblings->size; i++) {
		n += siblings->siblings[i]->size;
	}

	fl = sqalloc(sizeof(*fl));
	fl->size = n;
	fl->figures = sqalloc(fl->size * sizeof(*fl->figures));

	/* fill up */
	n = 0;
	for (i = 0; i < siblings->size; i++) {
		jsize = siblings->siblings[i]->size;
		f = siblings->siblings[i]->figures;
		for (j = 0; j < jsize; j++) {
			fl->figures[n] = &f[j];
			n++;
		}
	}

	/* Note: merging the many (small) ordered lists into the
	* flattened larger list instead of sorting the whole thing
	* would probably be slower.
	*/

	/* normalize (order) */
	normalizeFigureList(fl);

	return fl;
}


/** Search and sort helper [bsearch(), qsort(), etc].
*/
static int squarecmp1(void const* a, void const* b)
{
	square_t const* aa = (square_t const*)a;
	square_t const* bb = (square_t const*)b;

	return SQUARE_COMPARE(aa, bb);
}


/** Find a square (in an ordered array).
*/
static square_t* findSquare1(
	square_t* squares,
	int from,
	int to,
	square_t const* what,
	square_t** position)
{
	/* squares are ordered, do a fast binary search */
	return (square_t*)search(
		what,
		&squares[from],
		to - from,
		sizeof(*what),
		squarecmp1,
		(void**)position);
}


/** Compare two figures.
*/
static int figureCompare(figure_t const* a, figure_t const* b)
{
	int commonSize, i, sc;
	square_t const *ai, *bi;

	commonSize = (a->size <= b->size ? a->size : b->size);

	for (i = 0; i < commonSize; i++) {
		ai = &a->squares[i];
		bi = &b->squares[i];
		sc = SQUARE_COMPARE(ai, bi);
		if (sc != 0) {
			/* differ at some location */
			return sc;
		}
	}

	if (a->size != b->size) {
		/* one is larger */
		return (a->size > b->size ? 1 : -1);
	}

	/* they're the same */
	return 0;
}


/** Search and sort helper [bsearch(), qsort(), etc].
*/
static int figurecmp1(void const* a, void const* b)
{
	figure_t const* aa = (figure_t const*)a;
	figure_t const* bb = (figure_t const*)b;

	return figureCompare(aa, bb);
}


/** Search and sort helper [bsearch(), qsort(), etc].
*/
static int figurecmp2(void const* a, void const* b)
{
	figure_t const* aa = *(figure_t const**)a;
	figure_t const* bb = *(figure_t const**)b;

	return figureCompare(aa, bb);
}


/** Find a figure (in an ordered array).
*/
static figure_t* findFigure2(
	figure_t** figures,
	int from,
	int to,
	figure_t const* what,
	figure_t*** position)
{
	/* figures are ordered, do a fast binary search */
	return (figure_t*)search(
		&what,
		&figures[from],
		to - from,
		sizeof(what),
		figurecmp2,
		(void**)position);
}


/** Normalize a figure.  Destructive.
*/
static void normalizeFigure(figure_t* figure)
{
	square_t* s;
	int i, minX, minY;

	/* normalize offsets */

	/* find the minimas */

	/* catch empty case */
	if (figure->size == 0) {
		return;
	}

	s = &figure->squares[0];
	minX = s->x;
	minY = s->y;

	for (i = 1; i < figure->size; i++) {
		s = &figure->squares[i];
		if (s->x < minX) {
			minX = s->x;
		}
		if (s->y < minY) {
			minY = s->y;
		}
	}

	/* check if minima are already normalized */
	if (minX != 0 || minY != 0) {
		for (i = 0; i < figure->size; i++) {
			s = &figure->squares[i];
			s->x = (squareCoordinate_t)(s->x - minX);
			s->y = (squareCoordinate_t)(s->y - minY);
		}
	}

	/* normalize order (i.e.: sort) */
	sort(figure->squares, figure->size, sizeof(*figure->squares), squarecmp1);
}


/** Normalize a figure list.  Destructive.
*/
static void normalizeFigureList(figureList_t* figureList)
{
	/* normalize order (i.e.: sort) */
	sort(
		figureList->figures,
		figureList->size,
		sizeof(*figureList->figures),
		figurecmp2);
}


/** Normalize a sibling group.  Destructive.
*/
static void normalizeSiblingGroup(siblingGroup_t* siblingGroup)
{
	/* normalize order (i.e.: sort) */
	/* Note: [considering a sibling group carries only 1, 2, 4
	* or 8 elements] we are assuming that sort() [and qsort()]
	* are implementing small sorts (with up yo 4, possibly 8
	* elements) reasonably fast (i.e. not using a pivot, etc)
	* and are quite shortcutting sorts with element counts of
	* 0 or 1.
	*/
	sort(
		siblingGroup->figures,
		siblingGroup->size,
		sizeof(*siblingGroup->figures),
		figurecmp1);
}


/** Search and sort helper [bsearch(), qsort(), etc].
*/
static int siblinggroupcmp2(void const* a, void const* b)
{
	siblingGroup_t const* aa = *(siblingGroup_t const**)a;
	siblingGroup_t const* bb = *(siblingGroup_t const**)b;

	/* Sibling groups tend to be unique in their figures,
	* i.e. also in their first figure.
	*/
	return figureCompare(&aa->figures[0], &bb->figures[0]);
}


/** Find a sibling group (in an ordered array).
*/
static siblingGroup_t* findSiblingGroup2(
	siblingGroup_t** siblingGroups,
	int from,
	int to,
	siblingGroup_t const* what,
	siblingGroup_t*** position)
{
	/* sibling groups are ordered, do a fast binary search */
	return (siblingGroup_t*)search(
		&what,
		&siblingGroups[from],
		to - from,
		sizeof(what),
		siblinggroupcmp2,
		(void**)position);
}


/** Normalize a sibling list.  Destructive.
*/
static void normalizeSiblingList(siblingList_t* siblingList)
{
	/* normalize order (i.e.: sort) */
	sort(
		siblingList->siblings,
		siblingList->size,
		sizeof(*siblingList->siblings),
		siblinggroupcmp2);
}


/** Extended binary search.  A position is returned independant of the
* key being found.  Elements must be ordered and strictly ascending.
*/
static void* bsearchx(
	void const* key,
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b),
	void** position)
{
	char *low, *high, *mid; /* char* type for pointer arithmetc */
	int n, half, cmp;

	n = count;
	low = (char*)base; /* inclusive */
	high = &low[n * width]; /* exclusive */

	while (n > 0) {

		/* conditions: n > 0; high == &low[n * width] */

		half = n / 2; /* half < n [if n > 0] */
		mid = &low[half * width]; /* mid < high; mid == low if n == 1 */

		cmp = (*compare)(key, mid);
		if (cmp == 0) {
			/* Note: since elements must be strictly ascending, we're
			* not searching backwards for a 'first' match; this is
			* assumed to be the one and only match.
			*/
			if (position != NULL) {
				*position = mid;
			}
			return mid;
		}
		
		if (cmp < 0) {
			high = mid;
			n = half;
		} else { /* cmp > 0 */
			low = &mid[width];
			n -= half + 1;
		}
	}

	/* conditions: n == 0; high == &low[n * width] == low */

	if (position != NULL) {
		*position = high;
	}
	return NULL;
}


/** Search in an ordered array.  Has almost the same interface and is
* almost an abstraction of bsearch().
*/
static void* search(
	void const* key,
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b),
	void** position)
{
	return bsearchx(key, base, count, width, compare, position);
}


/** Sort.  Has the same interface and is an abstraction of qsort().
*/
static void sort(
	void* base,
	int count,
	int width,
	int (*compare)(void const* a, void const* b))
{
	qsort(base, count, width, compare);
}


/** Allocate memory.  Does not return if allocation fails.
*/
static void* sqalloc(int size)
{
	void* p;

	/* delegate to sqrealloc() */
	p = sqrealloc(NULL, size);

	/* initialize */
	memset(p, 0, size);

	return p;
}


/** Reallocate memory.  Does not return if (re-)allocation fails.
* sqrealloc(NULL,size) is identical to sqalloc(size).
*/
static void* sqrealloc(void* p, int size)
{
	p = realloc(p, size);
	if (p == NULL) {
		abort();
	}

	/* Note: can't initialize due to missing information about old size. */

	return p;
}


/** Deallocate (free) memory previously allocated through sqalloc() or
* sqrealloc().
*/
static void sqfree(void* p)
{
	if (p == NULL) {
		return;
	}

	free(p);
}


/** Reallocate memory, keeping power of two size.  Does not return if
* (re-)allocation fails.  sqreallocExponential() helps having less
* internal reallocations and copying around, i.e. results in a speed-up.
*/
static void* sqreallocExponential(void* p, int size)
{
	int potSize;

	potSize = 1;
	while (potSize < size) {
		potSize *= 2;
	}

	return sqrealloc(p, potSize);
}

