/* sqfig.h -- Square-Figure functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqfig.h,v 1.1 2003/03/07 16:46:43 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


#ifndef SQFIG_H
#define SQFIG_H


/* Note: the 'constness' of returned data structures is considered deep.
*/


/** A square coordinate.  Coordinates are signed rather than unsigned to
* allow the implementation of simpler rotate functions.  Coordinates are
* chars (rather than ints) to save space (the problem domain won't allow
* figures of a size larger than 127 [min(SCHAR_MAX,-SCHAR_MIN)] anyway).
*/
typedef signed char squareCoordinate_t;


/** A square.  An X an a Y coordinate.
*/
typedef struct
{
	squareCoordinate_t x;
	squareCoordinate_t y;
}
square_t;

/* mind the multiple argument evaluation! */
#define SQUARE_COPY(d, s) ((d)->x = (s)->x, (d)->y = (s)->y)
/* mind the possible multiple argument evaluation! */
#define SQUARE_EQUAL(a, b) ((a)->x == (b)->x && (a)->y == (b)->y)

/* mind the possible multiple argument evaluation! */
/* Y has precedence over X.  This may make the figures a bit easier to
* plot and keep their order more intuitive.  Note that print and plot
* routines may decide to switch coordinates (which is e.g. cooler to
* read in string notation.
*/
#define SQUARE_COMPARE(a, b) \
	((a)->y != (b)->y ? ((a)->y < (b)->y ? -1 : 1) : \
	((a)->x != (b)->x ? ((a)->x < (b)->x ? -1 : 1) : 0))


#define MAXFIGURESIZE 16

/** A figure.  An array of squares.  The squares are ordered.  The array
* has fixed size to allow simpler setup (e.g. when using automatic
* memory).
*/
typedef struct
{
	int size;
	square_t squares[MAXFIGURESIZE];
}
figure_t;

/* this needs stdlib.h */
#define FIGURE_COPY(d, s) (memcpy((d), (s), sizeof(*(d))))
/* this requires figureCompare() to be declared and implemented */
#define FIGURE_EQUAL(a, b) (figureCompare((a), (b)) == 0)


/** An ordered list of pointers to figures.
*/
typedef struct
{
	int size;
	figure_t** figures;
}
figureList_t;


/** A group of sibling figures, i.e. figures obtained by rotation and
* mirroring.  The figures are ordered.  The array has fixed size [8,
* which is the maximum possible, if all symmetry lacks] to allow simpler
* setup (e.g. when using automatic memory).  size is (depending on the
* figure's symmetry) 1, 2, 4 or 8 (i.e.: at least 1).
*/
typedef struct
{
	int size;
	figure_t figures[8];
}
siblingGroup_t;


/** An ordered list of pointers to sibling groups.
*/
typedef struct
{
	int size;
	siblingGroup_t** siblings;
}
siblingList_t;


/** A complete set of figures of a specific size.  Two representations
* are provided: the ordered set of siblings and its flattened version.
* The redundant flat list is hold for performant searching.
*/
typedef struct
{
	siblingList_t* siblings;
	figureList_t* figures;
}
figureSet_t;

#define FIGURE_COUNT(f) ((f)->figures->size)
#define DISTINCT_FIGURE_COUNT(f) ((f)->siblings->size)


/** Return a figure set; create it if necessary.
*/
figureSet_t const* getFigureSet(int size);


#endif

