/* sqfigdr.c -- Square-Figure driver program.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqfigdr.c,v 1.2 2003/03/17 09:42:08 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "sqfig.h"
#include "sqpr.h"


/* local stuff */
static void counts(int upperSize);
static void demo(int upperSize, int plotSize);
static int getNumericalArgument(char const* s);
static int numberDigit(char c);
static void plot(int argc, char** argv, int plotSize);
static void textDemo(int upperSize);
static void usage(void);


/** sqfigdr application starting point.
*/
int main(int argc, char** argv)
{
	int doTextDemo = 0, doDemo = 0, doCounts = 0;
	int plotSize = 3; /* default [note: don't use zero or one as default] */
	int upperSize = MAXFIGURESIZE; /* (large) default */
	int i;

	/* option parsing */
	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "-counts") == 0) {
			doCounts = 1;
			if (i + 1 < argc && numberDigit(argv[i + 1][0])) {
				upperSize = getNumericalArgument(argv[i + 1]);
				i++;
			}
		} else if (strcmp(argv[i], "-demo") == 0) {
			doDemo = 1;
			if (i + 1 < argc && numberDigit(argv[i + 1][0])) {
				upperSize = getNumericalArgument(argv[i + 1]);
				i++;
			}
		} else if (strcmp(argv[i], "-plotsize") == 0) {
			if (i + 1 >= argc) {
				usage();
			}
			plotSize = getNumericalArgument(argv[i + 1]);
			i++;
		} else if (strcmp(argv[i], "-textdemo") == 0) {
			doTextDemo = 1;
			if (i + 1 < argc && numberDigit(argv[i + 1][0])) {
				upperSize = getNumericalArgument(argv[i + 1]);
				i++;
			}

		} else if (argv[i][0] == '-' && numberDigit(argv[i][1])) {
			plotSize = getNumericalArgument(&argv[i][1]);

		/* usages, end of options */
		} else if (strcmp(argv[i], "-?") == 0) {
			usage();
		} else if (argv[i][0] == '-') {
			usage();
		} else {
			break;
		}
	}

	/* dispatch */
	if (doTextDemo) {
		textDemo(upperSize);
	} else if (doDemo) {
		demo(upperSize, plotSize);
	} else if (doCounts) {
		counts(upperSize);
	} else {
		if (i == argc) {
			/* default usage */
			demo(upperSize, plotSize);
		} else {
			plot(argc - i, &argv[i], plotSize);
		}
	}

	return 0;
}


/** Usage message.
*/
static void usage(void)
{
	printf(
		"usage: sqfig"
		" [-plotsize <size>|-<size>]"
		" [{-counts|-demo|-textdemo} [<uppersize>]]"
		" [<size>[.<number>[.<sibling>]] ...]"
		"\n"
	);
	exit(1);
}


/** Check for number digit.
*/
static int numberDigit(char c)
{
	/* note: own implementation to make things more robust (e.g. against
	* negative values) and to encapsulate
	*/

	/* note: ascii only */
	return (c >= '0' && c <= '9');
}


/** Get a numerical argument.  Calls usage() if the argument is
* non-numerical or negative.
*/
static int getNumericalArgument(char const* s)
{
	int n;
	char* p;

	n = strtol(s, &p, 10);
	if (p == s) {
		usage();
	}
	if (n < 0) {
		usage();
	}

	return n;
}


/** Demonstration, textual.
*/
static void textDemo(int upperSize)
{
	int size;
	figureSet_t const* fs;

	for (size = 0; size < upperSize; size++) {

		printf("---- figure size %d ----\n", size);

		fs = getFigureSet(size);

		printf("figureset:\n");
		printFigureSet(fs);
		printf("\n");

		printf("flat:\n");
		printFigureSetFlat(fs);
		printf("\n");

		printf("distincts:\n");
		printFigureSetDistincts(fs);
		printf("\n");
	}
}


/** Demonstration.
*/
static void demo(int upperSize, int plotSize)
{
	int size;
	figureSet_t const* fs;

	for (size = 0; size < upperSize; size++) {
		fs = getFigureSet(size);
		plotFigureSet(fs, plotSize);
		printf("\n\n\n");
	}
}


/** Figure counts.
*/
static void counts(int upperSize)
{
	int size, c, d;
	figureSet_t const* fs;

	for (size = 0; size < upperSize; size++) {

		printf("figure size %d: ", size);
		fs = getFigureSet(size);
		c = FIGURE_COUNT(fs);
		d = DISTINCT_FIGURE_COUNT(fs);
		printf("%d figures (%d distincts)\n", c, d);
	}
}


/** Plot Figures.
*/
static void plot(int argc, char** argv, int plotSize)
{
	int i, n;
	char *p, *q;
	int figureSize, figureNumber, siblingNumber;
	figureSet_t const* fs;

	/* plot string specified figures */
	n = 0;
	for (i = 0; i < argc; i++) {
		figureSize = figureNumber = siblingNumber = -1;

		p = argv[i];

		/* accept # notation */
		if (*p == '#') {
			p = &p[1];
		}

		/* parse figure size, must be >=1 */
		figureSize = strtol(p, &q, 10);
		if (q == p) {
			usage();
		}
		if (figureSize < 0) {
			usage();
		}

		/* parse optional figure number */
		if (q != NULL && *q == '.') {
			p = &q[1];
			figureNumber = strtol(p, &q, 10);
			/* specify from 1, internal from 0 */
			figureNumber--;

			/* parse optional sibling number */
			if (q != NULL && *q == '.') {
				p = &q[1];
				siblingNumber = strtol(p, &q, 10);
				/* specify from 1, internal from 0 */
				siblingNumber--;
			}
		}

		/* figure number must be specified if sibling number is */
		if (figureNumber < 0 && siblingNumber >= 0) {
			usage();
		}

		fs = getFigureSet(figureSize);

		/* skip out-of-bounds */
		if (figureNumber >= 0 && figureNumber >= fs->siblings->size) {
			continue;
		}
		if (
			figureNumber >= 0 &&
			siblingNumber >= 0 &&
			siblingNumber >= fs->siblings->siblings[figureNumber]->size
		) {
			continue;
		}

		/* separator */
		if (n > 0) {
			printf("\n\n");
		}

		if (siblingNumber >= 0) {
			plotFigure(
				&fs->siblings->siblings[figureNumber]->figures[siblingNumber],
				plotSize);
		} else if (figureNumber >= 0) {
			plotSiblingGroup(fs->siblings->siblings[figureNumber], plotSize);
		} else {
			plotFigureSet(fs, plotSize);
		}

		n++;
	}
}

