/* sqpr.c -- Square-Figure print functions.
**
** Copyright (C) 2003 Eric Laroche.  All rights reserved.
**
** @author Eric Laroche <laroche@lrdev.com>
** @version @(#)$Id: sqpr.c,v 1.1 2003/03/07 16:46:38 laroche Exp $
**
** This program is free software;
** you can redistribute it and/or modify it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
*/


/* implemented interface */
#include "sqpr.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/* local stuff */
static int plotBlanks(int n);
static int plotDash(int size, int inside);
static void plotFigureList(figureList_t const* figureList, int plotSize);
static int plotSome(char const* s, int n);
static int plotTerminator(int size, int inside);
static void printFigureList(figureList_t const* figureList);
static void printSquare(square_t const* square);


/** Print a figure set's figures, hierarchical.
*/
void printFigureSet(figureSet_t const* figureSet)
{
	int i;

	/* list/set specifier */
	printf("{");

	/* iterate over the figures */
	for (i = 0; i < figureSet->siblings->size; i++) {
		/* separator */
		if (i != 0) {
			printf(", ");
		}
		printSiblingGroup(figureSet->siblings->siblings[i]);
	}

	printf("}");
}


/** Print a figure set's figures, flat.
*/
void printFigureSetFlat(figureSet_t const* figureSet)
{
	printFigureList(figureSet->figures);
}


/** Print a figure set's distinct figures (omitting rotameres and
* mirrors).
*/
void printFigureSetDistincts(figureSet_t const* figureSet)
{
	int i;

	/* list/set specifier */
	printf("{");

	/* iterate over the figures */
	for (i = 0; i < figureSet->siblings->size; i++) {
		/* separator */
		if (i != 0) {
			printf(", ");
		}
		/* print first */
		printFigure(&figureSet->siblings->siblings[i]->figures[0]);
	}

	printf("}");
}


/** Print a figure.
*/
void printFigure(figure_t const* figure)
{
	int i;

	/* iterate over the squares */
	for (i = 0; i < figure->size; i++) {
		/* separator */
		if (i != 0) {
			printf(".");
		}
		printSquare(&figure->squares[i]);
	}
}


/** Print a figure list.
*/
static void printFigureList(figureList_t const* figureList)
{
	int i;

	/* list/set specifier */
	printf("{");

	/* iterate over the figures */
	for (i = 0; i < figureList->size; i++) {
		/* separator */
		if (i != 0) {
			printf(", ");
		}
		printFigure(figureList->figures[i]);
	}

	printf("}");
}


/** Print a figure set's figures, hierarchical.
*/
void printSiblingGroup(siblingGroup_t const* siblingGroup)
{
	figureList_t fl;
	figure_t* fs[8];
	int i;

	fl.size = 0;
	fl.figures = fs;

	/* iterate over the figures, convert data format */
	for (i = 0; i < siblingGroup->size; i++) {
		fl.figures[fl.size] = (figure_t*)&siblingGroup->figures[i];
		fl.size++;
	}

	printFigureList(&fl);
}


/** Print a square.
*/
static void printSquare(square_t const* square)
{
	printf("%d.%d",

#ifdef YOVERX
/* Y has precedence over X. */
		(int)square->x,
		(int)square->y
#else
/* X has precedence over Y (default).
* This is cooler to read in string notation.
*/
		(int)square->y,
		(int)square->x
#endif

	);
}


/** Plot a figure set's figures.
*/
void plotFigureSet(figureSet_t const* figureSet, int plotSize)
{
	int i;

	/* iterate over the figures */
	for (i = 0; i < figureSet->siblings->size; i++) {
		/* separator */
		if (i != 0) {
			printf("\n\n");
		}

		plotSiblingGroup(figureSet->siblings->siblings[i], plotSize);
	}
}


/** Plot a sibling group's figures.
*/
void plotSiblingGroup(siblingGroup_t const* siblingGroup, int plotSize)
{
	int i;
	figureList_t fl;
	figure_t* fs[8];

	/* iterate over the figures, convert data format */
	fl.size = 0;
	fl.figures = fs;
	for (i = 0; i < siblingGroup->size; i++) {
		fl.figures[fl.size] = (figure_t*)&siblingGroup->figures[i];
		fl.size++;
	}

	plotFigureList(&fl, plotSize);
}


/** Plot a figure.
*/
void plotFigure(figure_t const* figure, int plotSize)
{
	figureList_t fl;
	figure_t* fs[1];

	/* convert data format */
	fl.size = 0;
	fl.figures = fs;
	fl.figures[fl.size] = (figure_t*)figure;
	fl.size++;

	plotFigureList(&fl, plotSize);
}


/** Plot a figure list.
*/
static void plotFigureList(figureList_t const* figureList, int plotSize)
{
	int i, j, k, m, x, y, maxX, maxY, spaces;
	int interSpacing, realPlotSize;
	int needsTerminator;
	figure_t const* f;
	square_t const *s, *t, *u;
	int maxXCache[8];

	interSpacing = 2; /* [one more than 1] */
	if (plotSize > 0) {
		interSpacing++; /* one for the terminator */
	}

	realPlotSize = (plotSize != 0 ? plotSize : 1);

	/* [init max X cache; it is filled only later
	* to minimize code redundancy]
	*/
	for (i = 0; i < sizeof(maxXCache) / sizeof(*maxXCache); i++) {
		maxXCache[i] = -1;
	}

	/* find the maximal Y over all figures */
	maxY = 0; /* at least one of the squares is normalized to Y==0 */
	for (i = 0; i < figureList->size; i++) {
		f = figureList->figures[i];
		for (j = 0; j < f->size; j++) {
			s = &f->squares[j];
			if (s->y > maxY) {
				maxY = s->y;
			}
		}
	}

	/* render Y-lines, from maxY down to 0 (i.e.: plot large Y first),
	* -1 to rend the bottom line
	*/
	for (y = maxY; y >= -1; y--) {
		spaces = 0;
		needsTerminator = 0;

		/* plot edge (0) and inside (>0);
		* allow plotting with zero plot size but break on the y==-1 case
		*/
		for (i = 0;
			((i < plotSize || i == 0) && y >= 0) || (i == 0 && plotSize > 0);
			i++) {

			/* iterate over the figures */
			for (j = 0; j < figureList->size; j++) {
				f = figureList->figures[j];

				/* find this figure's maximum X */
				/* have a cache for the most probably used
				* (up to the 8 of a sibling group)
				*/
				if (
					j < sizeof(maxXCache) / sizeof(*maxXCache) &&
					maxXCache[j] != -1
				) {
					maxX = maxXCache[j];

				} else {
					maxX = 0;
					for (k = 0; k < f->size; k++) {
						s = &f->squares[k];
						if (s->x > maxX) {
							maxX = s->x;
						}
					}
					maxX++; /* exclusive */

					/* feed cache */
					if (j < sizeof(maxXCache) / sizeof(*maxXCache)) {
						maxXCache[j] = maxX;
					}
				}

				/* plot */
				x = 0;
				for (k = 0, m = 0;;) {

					/* search a Y-line,
					* skip unrelated lines
					*/
					t = NULL;
					for (; k < f->size; k++) {
						s = &f->squares[k];
						if (s->y == y) {
							t = s;
							break;
						}
					}

					/* search a Y+1-line (but only if not inside,
					* using common edges and not at the start line)
					*/
					u = NULL;
					if (i == 0 && plotSize > 0 && y != maxY) {
						for (; m < f->size; m++) {
							s = &f->squares[m];
							if (s->y == y + 1) {
								u = s;
								break;
							}
						}
					}

					/* end condition [no more elements found] */
					if (t == NULL && u == NULL) {
						break;
					}

					if (u == NULL) {
						s = t;
						k++;
					} else if (t == NULL) {
						s = u;
						m++;
					} else if (t->x < u->x) {
						s = t;
						k++;
					} else if (t->x > u->x) {
						s = u;
						m++;
					} else { /* t->x == u->x */
						s = t;
						k++;
						m++;
					}

					/* check for gaps */
					if (s->x > x) { /* x must be ordered! */
						spaces += (s->x - x) * realPlotSize;
						x = s->x;
					}

					/* print accumulated blanks
					* [either between the figures or intra-figure]
					*/
					if (spaces > 0) {
						if (needsTerminator) {
							/* [intra-figure case] */
							spaces -= plotTerminator(plotSize, i != 0);
							needsTerminator = 0;
						}
						spaces -= plotBlanks(spaces); /* i.e. spaces = 0; */
					}

					plotDash(plotSize, i != 0);
					x++;
					needsTerminator = 1;
				}

				if (needsTerminator) {
					/* [extra-figure case] */
					spaces -= plotTerminator(plotSize, i != 0);
					needsTerminator = 0;
				}
				if (maxX > x) {
					spaces += (maxX - x) * realPlotSize;
					x = maxX;
				}

				/* space to next figure */
				spaces += interSpacing;
			}

			printf("\n");
			spaces = 0;
		}
	}
}


/** Plot some blanks.  Returns the number of characters printed.
*/
static int plotBlanks(int n)
{
	return plotSome(" ", n);
}


/** Plot some dashes.  Returns the number of characters printed.
*/
static int plotDash(int size, int inside)
{
	int n;

	/* small special case */
	if (size == 0) {
		return plotSome("#", 1);
	} else {
		n = plotSome((inside ? "|" : "+"), 1);
		n += plotSome((inside ? " " : "-"), size - 1);
		return n;
	}
}


/** Plot a terminator.  The plotsize is needed to handle the special
* small case.  Returns the number of characters printed.
*/
static int plotTerminator(int size, int inside)
{
	if (size != 0) {
		return plotSome((inside ? "|" : "+"), 1);
	}

	return 0;
}


/** Plot some characters.  Returns the number of characters printed.
*/
static int plotSome(char const* s, int n)
{
	int i;

	for (i = 0; i < n; i++) {
		printf("%s", s);
	}

	return n * strlen(s);
}

